"""Parse Rule Expressions."""

from __future__ import annotations

import calendar
import datetime
import warnings
from typing import Final

from lark import Lark, Token, Transformer, v_args

from .datecalc import (
    days_relative_to,
    last_wd_of_month,
    wd_of_month,
    wd_relative_to,
)
from .model import Month, WeekDay

__all__ = ['rule_parser']


rule_grammar: Final = r"""
    rule: recurrence [ _IF condition _ELSE rule ]

    recurrence: offset_rule
        | weekday_rule
        | "(" rule ")"
        | literal
        | NEVER
        | NAME

    offset_rule: NUMBER unit preposition recurrence

    weekday_rule: owm_rule | wd_rule | lwd_rule

    wd_rule: _THE? [ordinal] weekday [NOT] preposition recurrence

    owm_rule: _THE? ordinal weekday month_hook

    lwd_rule: _THE? _LAST weekday month_hook

    ?month_hook: _OF month

    preposition: BEFORE | AFTER

    ?unit: DAYS | WEEKS

    literal: month NUMBER

    weekday: MONDAY
        | TUESDAY
        | WEDNESDAY
        | THURSDAY
        | FRIDAY
        | SATURDAY
        | SUNDAY

    month: JANUARY
        | FEBRUARY
        | MARCH
        | APRIL
        | MAY
        | JUNE
        | JULY
        | AUGUST
        | SEPTEMBER
        | OCTOBER
        | NOVEMBER
        | DECEMBER

    condition: or_condition

    or_condition: and_condition ["or"i or_condition]

    and_condition: simple_condition ["and"i and_condition]

    simple_condition: recurrence_condition
        | year_condition
        | TRUE
        | FALSE

    ?recurrence_condition: recur_ref _EXISTS -> exist_condition
        | recur_ref [NOT] _IN month -> month_condition
        | recur_ref _IS [NOT] (weekday | NEVER) -> wd_condition
        | recur_ref _IS [NOT] _SAME _AS recur_ref -> day_eq_condition
        | recur_ref _IS [NOT] preposition recur_ref -> day_prep_condition

    ?recur_ref: recurrence

    year_condition: _YEAR year_predicate

    year_predicate: ydiv_cond | ycmp_cond

    ydiv_cond: _IS [NOT] division
    ycmp_cond: _IS? [NOT] preposition NUMBER

    division: LEAP
        | NUMBER [ _MOD NUMBER] -> ymod_cond

    ?ordinal: FIRST | SECOND | THIRD | FOURTH | TH

    MONDAY.7: /mo(n(day)?)?/i
    TUESDAY.7: /tu(e(sday)?)?/i
    WEDNESDAY.7: /we(d(nesday)?)?/i
    THURSDAY.7: /th(u(rsday))?/i
    FRIDAY.7: /fr(i(day)?)?/i
    SATURDAY.7: /sa(t(urday)?)?/i
    SUNDAY.7: /su(n(day)?)?/i

    JANUARY.9: /jan(uary)?/i
    FEBRUARY.9: /feb(ruary)?/i
    MARCH.9: /mar(ch)?/i
    APRIL.9: /apr(il)?/i
    MAY.9: /may/i
    JUNE.9: /june?/i
    JULY.9: /july?/i
    AUGUST.9: /aug(ust)?/i
    SEPTEMBER.9: /sep(temb(er|re))?/i
    OCTOBER.9: /oct(ob(er|re))?/i
    NOVEMBER.9: /nov(emb(er|re))?/i
    DECEMBER.9: /dec(emb(er|re))?/i

    TRUE.9: /true/i
    FALSE.9: /false/i

    FIRST.9: "first"i
    SECOND.9: "second"i
    THIRD.9: "third"i
    FOURTH.9: "fourth"i
    _LAST.9: "last"i
    _EXISTS.9: "exists"i
    _OF.9: /of/i
    _IF.9: "if"i
    _ELSE.9: "else"i
    NEVER.9: "never"i
    _IS.9: "is"i
    _IN.9: "in"i
    LEAP.9: "leap"i
    _MOD.9: "mod"i
    _SAME.9: "same"i
    _AS.9: "as"i
    _THE.9: "the"i
    BEFORE.9: "before"i
    AFTER.9: "after"i
    _YEAR.9: "year"i
    DAYS.9: /days?/i
    WEEKS.9: /weeks?/i
    NOT.9: "not"i

    TH.9: /[2-9]?(1st|2nd|3rd|([4-90]th))|1[0-9]th/i
    NAME.0: /[a-zA-Z]([-_.]?[a-zA-Z0-9])*/
    NUMBER.0: /[0-9]+/

    %ignore /[ \t\n\r]+/
    """


@v_args(inline=True)
class RuleEvaluator(Transformer):
    """Evaluate recurrence rules for a given year.

    Properties:
    -----------
    - funcs
        a dictionary of precomputed dates
    - year
        the year for which new dates are computed
    """

    def __init__(
        self,
        funcs: dict[str, datetime.date | None],
        year: int,
    ) -> None:
        self.funcs: dict[str, datetime.date | None] = funcs
        self.year: int = year

    def rule(
        self,
        t_value: datetime.date | None,
        condition: bool | None,
        f_value: datetime.date | None,
    ) -> datetime.date | None:
        """Evaluate an optional conditional expression."""
        if (condition is None) or condition:
            return t_value
        return f_value

    def weekday_rule(
        self,
        rec: datetime.date | None,
    ) -> datetime.date | None:
        """Translate a weekday rule."""
        return rec

    def preposition(self, the_prep: int) -> int:
        """Return the actual preposition."""
        return the_prep

    def recurrence(self, rd: datetime.date | None) -> datetime.date | None:
        return rd

    def simple_condition(self, cond: bool) -> bool:
        """Evaluate condition."""
        return cond

    def condition(self, cond: bool) -> bool:
        """Evaluate condition."""
        return cond

    def literal(self, month: Month, day: int) -> datetime.date | None:
        """Convert literal to date."""
        try:
            return datetime.date(self.year, month.value, day)
        except ValueError:
            lit_str = f'{self.year}/{month}/{day}'
            warnings.warn(
                'Date literal cannot be converted: ' + lit_str,
                stacklevel=2,
            )
            return None

    def weekday(self, token) -> WeekDay:
        """Parse a weekday."""
        return WeekDay[token.type]

    def month(self, token: Token) -> Month:
        """Parse a weekday."""
        return Month[token.type]

    def NAME(self, token: Token) -> datetime.date | None:  # noqa: N802
        """Lookup name."""
        name = token.value
        if name not in self.funcs:
            warnings.warn(
                f'Unknown date function {name} referenced.',
                stacklevel=2,
            )
            return None
        return self.funcs[name]

    def NEVER(self, token: Token) -> None:  # noqa: N802
        """Translate ``NEVER`` to ``None``."""

    def NUMBER(self, token: Token) -> int:  # noqa: N802
        """Interpret nimber."""
        return int(token.value)

    def offset_rule(
        self,
        number: int,
        unit: int,
        preposition: int,
        recurrence: datetime.date | None,
    ) -> datetime.date | None:
        """Translate offset rules."""
        if recurrence is None:
            return None
        num_days = number * unit * preposition
        return days_relative_to(recurrence, num_days)

    def owm_rule(
        self,
        ordinal: int,
        week_day: WeekDay,
        hook: Month,
    ) -> datetime.date | None:
        """Translate weekday-of-month rule."""
        return wd_of_month(self.year, hook, ordinal, week_day)

    def wd_rule(
        self,
        ordinal: int | None,
        week_day: WeekDay,
        neg: Token | None,
        preposition: int,
        recurrence: datetime.date | None,
    ) -> datetime.date | None:
        """Translate weekday-relative-to rule."""
        if recurrence is None:
            return None
        include_start: bool = neg is not None
        direction = -preposition if include_start else preposition
        return days_relative_to(
            wd_relative_to(
                recurrence,
                week_day,
                direction,
                include_start,
            ),
            direction * 7 * (ordinal - 1 if ordinal else 0),
        )

    def lwd_rule(
        self,
        week_day: WeekDay,
        month: Month,
    ) -> datetime.date | None:
        """Compute last of a week day of a month."""
        return last_wd_of_month(self.year, month, week_day)

    def wd_condition(
        self,
        recur_ref: datetime.date | None,
        not_tok: Token | None,
        week_day: WeekDay | None,
    ) -> bool:
        """Evaluate weekday condition."""
        if week_day is None:
            # recur_ref IS [NOT] NEVER
            return _negate(not_tok, recur_ref is None)
        if recur_ref is None:
            return False
        return _negate(not_tok, recur_ref.weekday() == week_day.value)

    def day_eq_condition(
        self,
        recur_ref: datetime.date | None,
        not_tok: Token | None,
        recur_ref_2: datetime.date | None,
    ) -> bool:
        """Evaluate day equality condition."""
        if recur_ref is None or recur_ref_2 is None:
            return False
        return _negate(not_tok, recur_ref == recur_ref_2)

    def day_prep_condition(
        self,
        recur_ref: datetime.date | None,
        not_tok: Token | None,
        preposition: int,
        recur_ref_2: datetime.date | None,
    ) -> bool:
        """Evaluate day preposition condition."""
        if recur_ref is None or recur_ref_2 is None:
            return False
        diff = preposition * (recur_ref - recur_ref_2).days
        return _negate(not_tok, diff > 0)

    def month_condition(
        self,
        recur_ref: datetime.date | None,
        not_tok: Token | None,
        month: Month,
    ) -> bool:
        """Check whether a date is in the given month."""
        if recur_ref is None:
            return False
        m_cond = month.value == recur_ref.month
        y_cond = self.year == recur_ref.year
        return _negate(not_tok, m_cond and y_cond)

    def exist_condition(self, recurrence: datetime.date | None) -> bool:
        """Check a date for existence."""
        return recurrence is not None

    def and_condition(self, cond1: bool, cond2: bool) -> bool:
        """Evaluate an and-condition."""
        if cond2 is None:
            return cond1
        return cond1 and cond2

    def or_condition(self, cond1: bool, cond2: bool) -> bool:
        """Evaluate an or-condition."""
        if cond2 is None:
            return cond1
        return cond1 or cond2

    def BEFORE(self, token: Token) -> int:  # noqa: N802
        """Translate ``BEFORE`` to -1."""
        return -1

    def AFTER(self, token: Token) -> int:  # noqa: N802
        """Translate ``AFTER`` to +1."""
        return 1

    def DAYS(self, token: Token) -> int:  # noqa: N802
        """Translate ``DAYS`` to +1."""
        return 1

    def WEEKS(self, token: Token) -> int:  # noqa: N802
        """Translate ``WEEKS`` to 7."""
        return 7

    def FIRST(self, token: Token) -> int:  # noqa: N802
        """Translate ordinal to int."""
        return 1

    def SECOND(self, token: Token) -> int:  # noqa: N802
        """Translate ordinal to int."""
        return 2

    def THIRD(self, token: Token) -> int:  # noqa: N802
        """Translate ordinal to int."""
        return 3

    def FOURTH(self, token: Token) -> int:  # noqa: N802
        """Translate ordinal to int."""
        return 4

    def TH(self, token: Token) -> int:  # noqa: N802
        """Translate ordinal to int."""
        result: int = 0
        for dig in token.value:
            if dig >= '0' and dig <= '9':
                result *= 10
                result += ord(dig) - ord('0')
        return result

    def TRUE(self, token: Token) -> bool:  # noqa: N802
        """Transform boolean literal."""
        return True

    def FALSE(self, token: Token) -> bool:  # noqa: N802
        """Transform boolean literal."""
        return False

    def year_condition(self, year_predicate: bool) -> bool:
        """Evaluate a year condition."""
        return year_predicate

    def year_predicate(self, cond: bool) -> bool:
        """Evaluate year condition."""
        return cond

    def ydiv_cond(self, not_tok: Token | None, cond: bool) -> bool:
        """Evaluate division like conditions."""
        return _negate(not_tok, cond)

    def ycmp_cond(
        self,
        not_tok: Token | None,
        preposition: int,
        number: int,
    ) -> bool:
        """Compare year number."""
        return _negate(not_tok, (self.year - number) * preposition > 0)

    def division(self, cond: bool) -> bool:
        """Evaluate division rule."""
        return cond

    def LEAP(self, _: Token) -> bool:
        """Check whether year is a leap year."""
        return calendar.isleap(self.year)

    def ymod_cond(self, rem: int, divi: int | None) -> bool:
        """Check year in nodular arithmetic."""
        return rem == (self.year if not divi else (self.year % divi))


def _negate(not_tok: Token | None, cond: bool) -> bool:
    """Negate rhe condition if the NOT token is present."""
    return cond == (not_tok is None)


def rule_parser(
    year: int,
    funcs: dict[str, datetime.date | None] | None = None,
) -> Lark:
    """Generate rule parser.

    Arguments
    ---------
    year : int
        The year for which new dates are computed.
    funcs : dict[str, datetime.date | None] | None, optional
        A dictionary of precomputed dates.

    Return
    ------
    Lark
        A ``Lark`` parser instance.

    Example
    -------
    >>> from annual.ruleparser import rule_parser
    >>> rule_parser(2024).parse('Sunday after may 1')
    datetime.date(2024, 5, 5)

    You can use the ``funcs`` argument to supply precomputed dates,
    which can be referred to in the given expression.

    >>> from annual.ruleparser import rule_parser
    >>> year = 2024
    >>> funcs = {'easter': datetime.date(year, 3, 31)}
    >>> rule_parser(year, funcs).parse('49 days after easter')
    datetime.date(2024, 5, 19)
    """
    return Lark(
        rule_grammar,
        start='rule',
        parser='lalr',
        transformer=RuleEvaluator(funcs if funcs else {}, year),
    )



"""Decorators for date functions.

This module implements decorators for registering functions
at a central registry
of functions computing a single date per year and
of generators computing a list of events for a given year.
"""

import datetime
from collections.abc import Iterator
from functools import wraps
from typing import Callable, TypeAlias, no_type_check

__all__ = [
    'DateFunction',
    'DateIterator',
    'date_function',
    # 'date_generator',
]

MaybeDate: TypeAlias = datetime.date | None
DateFunction: TypeAlias = Callable[[int], MaybeDate]
DateIterator: TypeAlias = Callable[
    [int],
    Iterator[tuple[str, MaybeDate]],
]


def date_function(
    name: str | None = None,
) -> Callable[[DateFunction], DateFunction]:
    """Mark a date function to be made available in the parser.

    The function name is used in the registry if no name is
    passed explicitly.

    Parameters
    ----------
    name : str | None
        the name to be used in the registry (optional)

    Return
    ------
    Callable[[DateFunction], DateFunction]
        the actual decortor
    """

    def decorator_func(date_func: DateFunction) -> DateFunction:
        """Add the wrapper."""

        @wraps(date_func)
        def wrapper(year: int) -> datetime.date | None:
            """Wrap the function.."""
            return date_func(year)

        if name:
            wrapper.__name__ = name
        mark_decorator(wrapper, date_function.__name__)
        return wrapper

    return decorator_func


@no_type_check
def mark_decorator(
    wrapper: Callable[[int], MaybeDate],
    decorator: str,
) -> None:
    """Add a ``__decorator__`` attribute to a given ``wrapper`` function.

    Parameters
    ----------
    wrapper : Callable[[int], MaybeDate]
        the wrapper function

    decorator : str
        the value for the ``__decorator__`` attribute
    """
    wrapper.__decorator__ = decorator



"""Implementation of a registry for date functions and iterators."""

import datetime
import importlib
from collections.abc import Sequence
from importlib.metadata import entry_points
from typing import cast

from .decorators import DateFunction

__all__ = ['FunctionRegistry']


class FunctionRegistry:
    """Registry for date functions and iterators.

    Parameters
    ----------
    auto_plugins : bool
        determines whether plugins should be loaded upon initialization
        (optional, default = ``True``)
    """

    def __init__(self, auto_plugins: bool = True) -> None:
        self._date_functions: dict[str, DateFunction] = {}
        if auto_plugins:
            self.add_from_plugins()

    def add_from_plugins(
        self,
        exclude: Sequence[str] = (),
        only: Sequence[str] = (),
    ) -> None:
        """Add date functions and iterators from plugins.

        Parameters
        ----------
        exclude : Sequence[str], optional
            list of plugins not to be added
        only : Sequence[str], optional
            list of plugins to be added exclusively
        """
        discovered_plugins = entry_points(group='annual')
        for entry_point in discovered_plugins:
            if entry_point.name in exclude:
                continue
            if only and entry_point.name not in only:
                continue
            self.add_from_module(entry_point.value)

    def add_from_module(self, module_name: str) -> None:
        """Add date functions and iterators from a given module.

        Parameters
        ----------
        module_name : str
            the name of the module to be scanned for date functions
        """
        module = importlib.import_module(module_name)
        for name in dir(module):
            obj = getattr(module, name)
            if hasattr(obj, '__decorator__'):
                match obj.__decorator__:
                    case 'date_function':
                        self.add_date_function(cast(DateFunction, obj))

    def add_date_function(self, date_function: DateFunction) -> None:
        """Add the given date function.

        Parameters
        ----------
        date_function: DateFunction
            the date function to be added
        """
        self._date_functions[date_function.__name__] = date_function

    def evaluate(self, year: int) -> dict[str, datetime.date | None]:
        """Evaluate all registered functions for the given year.

        Parameters
        ----------
        year : int
            the year for which the functions are evaluated

        Return
        ------
        dict[str, datetime.date | None]
            a mapping between function names and their results
        """
        result: dict[str, datetime.date | None] = {}
        for name, date_function in self._date_functions.items():
            result[name] = date_function(year)
        return result



"""Annual date functions visible here are discovered automatically."""

from .easter_funcs import easter, easter_julian, easter_orthodox

__all__ = [
    'easter',
    'easter_julian',
    'easter_orthodox',
]



"""Data structures and enumerations."""

from __future__ import annotations

from enum import Enum

__all__ = ['Month', 'WeekDay']


class Month(Enum):
    """Enumerate the months of the year."""

    JANUARY = 1
    FEBRUARY = 2
    MARCH = 3
    APRIL = 4
    MAY = 5
    JUNE = 6
    JULY = 7
    AUGUST = 8
    SEPTEMBER = 9
    OCTOBER = 10
    NOVEMBER = 11
    DECEMBER = 12


class WeekDay(Enum):
    """Enumerate the days of the week."""

    MONDAY = 0
    TUESDAY = 1
    WEDNESDAY = 2
    THURSDAY = 3
    FRIDAY = 4
    SATURDAY = 5
    SUNDAY = 6



"""Easter date clculrion routines.

This module computes the easter date for any given year.

For implementation details see the `GM Arts`_ site on
easter algorithms.

.. _GM Arts: http://dates.gmarts.org/eastalg.htm
"""

import datetime

from ..decorators import date_function

__all__ = ['easter', 'easter_orthodox', 'easter_julian']


@date_function()
def easter(year: int) -> datetime.date | None:
    """Calculate the easter date for Western churches.

    This method computes the easter date according to
    the revised method in the Gregorian calendar.
    The method is valid for years between 1583 and 4099.

    Parameters
    ----------
    year : int
        the year in the range from 1583 to 4099

    Return
    ------
    datetime.date | None
        the easter date for the given year or ``None``
        if the method does not apply to the given year
    """
    if year < 1583 or year > 4099:
        return None

    pfm = paschal_full_moon(year)
    day = find_next_sunday(year, pfm, True)
    return easter_day_to_date(year, day)


@date_function()
def easter_orthodox(year: int) -> datetime.date | None:
    """Calculate the easter date for Eastern churches.

    This method computes the easter date according to
    the original Julian method for a Gregorian date.
    The method is valid for years between 1583 and 4099.

    Parameters
    ----------
    year : int
        the year in the range from 1583 to 4099

    Return
    ------
    datetime.date | None
        the easter date for the given year or ``None``
        if the method does not apply to the given year
    """
    if year < 1583 or year > 4099:
        return None
    golden = year % 19
    pfm = ((225 - 11 * golden) % 30) + 21
    day = find_next_sunday(year, pfm, False)
    day += julian_easter_to_gregorian_offset(year)
    return easter_day_to_date(year, day)


@date_function()
def easter_julian(year: int) -> datetime.date | None:
    """Calculate the Julian easter date.

    This method computes the easter date according to
    the original Julian method without conversion to
    Gregorian calendar.
    The method is valid for years after 326.

    Parameters
    ----------
    year : int
        the year, which must be after 326.

    Return
    ------
    datetime.date | None
        the easter date for the given year or ``None``
        if the method does not apply to the given year
    """
    if year < 326:
        return None
    golden = year % 19
    pfm = ((225 - 11 * golden) % 30) + 21
    day = find_next_sunday(year, pfm, False)
    return easter_day_to_date(year, day)


def paschal_full_moon(year: int) -> int:
    """Calculate Paschal Full Moon (PFM) date.

    Parameters
    ----------
    year : int
        the year

    Return
    ------
    int
        a number indicating the PFM date counting from 20 March
    """
    century = year // 100
    golden = year % 19

    temp = (century - 15) // 2 + 202 - 11 * golden
    if century in (21, 24, 25, 27, 28, 29, 30, 31, 32, 34, 35, 38):
        temp -= 1
    elif century in (33, 36, 37, 39, 40):
        temp -= 2
    temp %= 30

    if temp == 29 or (temp == 28 and golden > 10):
        return temp + 20
    return temp + 21


def find_next_sunday(year: int, pfm: int, is_western: bool) -> int:
    """Compute the next Sunday after paschalfull moon.

    Parameters
    ----------
    year : int
        the year of the resulting date
    pfm : int
        the paschal full moon
    is_western : bool
        indicate whether the Western (revised) method is used

    Return
    ------
    int
        the day in March of the Sunday after PFM, where 32 March = 1 April
    """
    term_b = (pfm - 19) % 7
    term_c = (40 - year // 100) % (4 if is_western else 7)
    if is_western:
        if term_c == 3:
            term_c += 1
        if term_c > 1:
            term_c += 1
    temp = year % 100
    term_d = (temp + temp // 4) % 7
    return pfm + ((20 - term_b - term_c - term_d) % 7) + 1


def julian_easter_to_gregorian_offset(year: int) -> int:
    """Compute the offset of the Julian easter date to Gregorian.

    Parameters
    ----------
    year : int
        the year

    Return
    ------
    int
        the number of dys to add to the Julian date
    """
    century = year // 100
    # 10 days were 'skipped' in the Gregorian calendar from 5-14 Oct 1582
    skip = 10
    # Only 1 in every 4 century years are leap years in the Gregorian
    # calendar (every century is a leap year in the Julian calendar)
    if century <= 16:
        return skip
    century -= 16
    return skip + century - (century // 4)


def easter_day_to_date(year: int, day: int) -> datetime.date:
    """Convert days after March 1st to a proper date.

    Parameters
    ----------
    year : int
        the year of the resulting date
    day : int
        the number of days counting from and including March 1st

    Return
    ------
    datetime.date
        a date in March, April or May
    """
    if day > 61:
        return datetime.date(year, 5, day - 61)
    if day > 31:
        return datetime.date(year, 4, day - 31)
    return datetime.date(year, 3, day)



"""Date calculations for the parser."""

from __future__ import annotations

import datetime

from .model import Month, WeekDay

__all__ = [
    'days_relative_to',
    'last_wd_of_month',
    'wd_of_month',
    'wd_relative_to',
]


def wd_of_month(
    year: int,
    month: Month,
    ordinal: int,
    week_day: WeekDay,
) -> datetime.date | None:
    """Compute the n-th occurrence of a weekday in a month.

    Ensures that the result is in the given ``year`` and ``month``.
    If the request cannot be fulfilled, ``None`` is returned.

    Parameters
    ----------
    year : int
        the (full four-digit) `year` for which the date should be computed
    month : Month
        the month for which the date is to be computed
    ordinal : int
        if set to 1 (resp. 2, 3, ...) the first (second, third, ...)
        occurrence of the given weekday is computed,
    week_day : WeekDay
        the weekday of the result

    Returns
    -------
    datetime.date | None
        The requested date, or ``None`` if no such date exists
    """
    first = datetime.date(year, month.value, 1)
    offs = (week_day.value - first.weekday() + 7) % 7
    result = days_relative_to(first, offs + (ordinal - 1) * 7)
    if result.year == year and result.month == month.value:
        return result
    return None


def last_wd_of_month(
    year: int,
    month: Month,
    week_day: WeekDay,
) -> datetime.date:
    """Compute the last occurrence of a weekday in a month."""
    target_year = year + 1 if month == Month.DECEMBER else year
    target_month = (month.value % 12) + 1
    first = datetime.date(target_year, target_month, 1)
    return wd_relative_to(first, week_day, -1, False)


def wd_relative_to(
    recurrence: datetime.date,
    week_day: WeekDay,
    direction: int,
    include_start: bool,
) -> datetime.date:
    """Compute when a weekday occurs relative to a given date.

    Parameters
    ----------
    recurrence : datetime.date
        the start date relative to which the result is computed
    week_day : WeekDay
        the weekday of rhe target date
    direction : int
        must be 1 to compute a date after the start date or -1
        to compute a date before
    include_start : bool
        skip one week if the week day of the start date matches

    Returns
    -------
    datetime.date
        the first occurrence of the given weekday before or after
        the given date
    """
    delta = (week_day.value - recurrence.weekday()) % 7
    if delta == 0:
        if include_start:
            return recurrence
        delta = 7 if direction > 0 else -7
    elif direction < 0:
        delta -= 7
    return days_relative_to(recurrence, delta)


def days_relative_to(
    recurrence: datetime.date,
    num_days: int,
) -> datetime.date:
    """Add given number of days to the given date.

    Parameters
    ----------
    recurrence : datetime.date
        start date to which days are added
    num_days : int
        number of days to add

    Returns
    -------
    datetime.date
        the `recurrence` date plus the given number of days
    """
    return recurrence + datetime.timedelta(days=num_days)
